from typing import List

from marshmallow import fields

from ..utils.facet_helpers import searchable
from .datatypes import DataType, Import


class NumberDataType(DataType):
    def marshmallow_validators(self):
        validators = []
        ranges = {}
        for param, schema in (
            ("min", "minimumExclusive"),
            ("max", "maximumExclusive"),
            ("min_inclusive", "minimum"),
            ("max_inclusive", "maximum"),
        ):
            if schema in self.definition:
                ranges[param] = self.definition[schema]

        if ranges:
            params = ", ".join(f"{k}={v}" for k, v in ranges.items())
            validators.append(f"ma_validate.Range({params})")

        return validators

    def facet(self, key, definition={}, props_num=None, create=True):
        if not searchable(definition, create):
            return False
        key = definition.get("key", key)
        field = definition.get("field", "TermsFacet(field = ")
        facet_def = {"path": key, "class": field}
        if "field" in definition:
            facet_def["defined_class"] = True
        return facet_def


class IntegerDataType(NumberDataType):
    marshmallow_field = "ma_fields.Integer"
    schema_type = "integer"
    model_type = "integer"

    class ModelSchema(DataType.ModelSchema):
        minimum = fields.Integer(required=False)
        exclusiveMinimum = fields.Integer(required=False)
        maximum = fields.Integer(required=False)
        exclusiveMaximum = fields.Integer(required=False)


class FloatDataType(NumberDataType):
    marshmallow_field = "ma_fields.Float"
    schema_type = "number"
    model_type = "float"

    class ModelSchema(DataType.ModelSchema):
        minimum = fields.Float(required=False)
        exclusiveMinimum = fields.Float(required=False)
        maximum = fields.Float(required=False)
        exclusiveMaximum = fields.Float(required=False)


class DoubleDataType(NumberDataType):
    marshmallow_field = "ma_fields.Double"
    schema_type = "number"
    model_type = "double"

    class ModelSchema(DataType.ModelSchema):
        minimum = fields.Float(required=False)
        exclusiveMinimum = fields.Float(required=False)
        maximum = fields.Float(required=False)
        exclusiveMaximum = fields.Float(required=False)


class BooleanDataType(DataType):
    marshmallow_field = "ma_fields.Boolean"
    model_type = "boolean"

    def facet(self, key, definition={}, props_num=None, create=True):
        if not searchable(definition, create):
            return False
        key = definition.get("key", key)
        field = definition.get("field", "TermsFacet(field = ")
        facet_def = {"path": key, "class": field}
        if "field" in definition:
            facet_def["defined_class"] = True
        return facet_def



import keyword
import re


def convert_name_to_python(name):
    # special case :)

    name = name.replace("@v", "_version")
    name = name.replace("id", "_id")
    # Replace any spaces or special characters in the string with an underscore
    identifier = re.sub(r"[^\w\s]", "_", name)
    identifier = re.sub(r"\s+", "_", identifier)

    # Ensure that the first character of the identifier is a letter or an underscore
    if identifier[0].isdigit():
        identifier = "_" + identifier

    # Ensure that the resulting identifier is not a reserved keyword in Python
    if keyword.iskeyword(identifier):
        identifier = identifier + "_"

    return identifier


def capitalize(s):
    return s[0].capitalize() + s[1:]


def convert_name_to_python_class(name):
    class_name = "".join([capitalize(word) for word in name.split("-")])
    class_name = "".join([capitalize(word) for word in class_name.split("_")])

    # Replace any spaces or special characters in the string with an underscore
    class_name = re.sub(r"[^\w\s]", "", class_name)
    class_name = re.sub(r"\s+", "_", class_name)

    return class_name



import os

from oarepo_model_builder.entrypoints import create_builder_from_entrypoints, load_model
from oarepo_model_builder.fs import InMemoryFileSystem


def test_no_dependencies():
    data = build()
    print(data)
    assert data.startswith("[metadata]")


def test_runtime_dependencies():
    data = build({"runtime-dependencies": {"test": "1.0.0"}})
    print(data)
    assert "install_requires =" in data
    assert """test>=1.0.0""" in data
    assert data.index("install_requires") < data.index("test>=1.0.0")


def test_dev_dependencies():
    data = build({"dev-dependencies": {"test": "1.0.0"}})
    assert "[options.extras_require]" in data
    assert "devs =" in data
    assert """test>=1.0.0""" in data
    assert data.index("devs =") < data.index("test>=1.0.0")


def build(kwargs={}):
    schema = load_model(
        "test.yaml",
        "test",
        model_content={"model": {}, **kwargs},
        isort=False,
        black=False,
    )
    filesystem = InMemoryFileSystem()
    builder = create_builder_from_entrypoints(filesystem=filesystem)
    builder.skip_schema_validation = True
    builder.build(schema, "")
    data = builder.filesystem.open("setup.cfg").read().strip()
    return data



from oarepo_model_builder.builders.jsonschema import JSONSchemaBuilder
from oarepo_model_builder.builders.mapping import MappingBuilder
from oarepo_model_builder.invenio.invenio_record_schema import (
    InvenioRecordSchemaBuilder,
)
from oarepo_model_builder.invenio.invenio_script_sample_data import (
    InvenioScriptSampleDataBuilder,
)
from oarepo_model_builder.property_preprocessors import PropertyPreprocessor, process
from oarepo_model_builder.stack import ModelBuilderStack
from oarepo_model_builder.utils.deepmerge import deepmerge


class EnumPreprocessor(PropertyPreprocessor):
    TYPE = "enum"

    @process(
        model_builder=InvenioRecordSchemaBuilder,
        path="/properties/*",
        condition=lambda current, stack: stack.top.schema_element_type == "property",
    )
    def modify_enum_marshmallow(self, data, stack: ModelBuilderStack, **kwargs):
        if "enum" not in data:
            return data

        # TODO: support other datatypes than string
        alternatives = [f'"{x}"' for x in data["enum"]]

        deepmerge(
            data.setdefault("marshmallow", {}),
            {"validators": [f'ma_valid.OneOf([{", ".join(alternatives)}])']},
        )
        return data

    @process(
        model_builder=JSONSchemaBuilder,
        path="/properties/*",
        condition=lambda current, stack: stack.top.schema_element_type == "property",
    )
    def modify_enum_jsonschema(self, data, stack: ModelBuilderStack, **kwargs):
        if "enum" not in data:
            return data

        deepmerge(data.setdefault("jsonschema", {}), {"enum": data["enum"]})
        return data

    @process(
        model_builder=InvenioScriptSampleDataBuilder,
        path="/properties/*",
        condition=lambda current, stack: stack.top.schema_element_type == "property",
    )
    def modify_enum_sampledata(self, data, stack: ModelBuilderStack, **kwargs):
        if "enum" not in data:
            return data
        if "sample" in data:
            return data
        data["sample"] = data["enum"]
        return data



import json
import logging
from hashlib import sha256

from marshmallow import fields

from oarepo_model_builder.utils.camelcase import camel_case
from oarepo_model_builder.utils.jinja import split_package_name
from oarepo_model_builder.utils.python_name import convert_name_to_python_class
from oarepo_model_builder.validation import InvalidModelException, model_validator

from ..utils.facet_helpers import searchable
from .datatypes import DataType

log = logging.getLogger("datatypes")


class ObjectDataType(DataType):
    schema_type = "object"
    mapping_type = "object"
    marshmallow_field = "ma_fields.Nested"
    model_type = "object"

    class ModelSchema(DataType.ModelSchema):
        properties = fields.Nested(
            lambda: model_validator.validator_class("properties", strict=False)()
        )

    def marshmallow(self, **extras):
        ret = super().marshmallow(**extras)
        schema_class = ret.get("schema-class", None)
        if not schema_class:
            if self.stack.top.schema_element_type == "items":
                schema_class_base = self.stack[-2].key + "Item"
            else:
                schema_class_base = self.key
            schema_class = convert_name_to_python_class(schema_class_base) + "Schema"

        package_name = split_package_name(self.model.record_schema_class)

        schema_class = self._get_class_name(package_name, schema_class)

        fingerprint = sha256(
            json.dumps(
                self.definition, sort_keys=True, default=lambda x: repr(x)
            ).encode("utf-8")
        ).hexdigest()

        if "known-classes" not in self.model:
            self.model.known_classes = {}

        schema_class = self._find_unique_schema_class(
            self.model.known_classes, schema_class, fingerprint
        )
        log.debug(
            "%s: fp %s, schema class %s", self.stack.path, fingerprint, schema_class
        )

        self.model.known_classes[schema_class] = fingerprint

        ret["schema-class"] = schema_class

        return ret

    def _find_unique_schema_class(self, known_classes, schema_class, fingerprint):
        if schema_class in known_classes:
            # reuse class with the same fingerprint
            if fingerprint != known_classes[schema_class]:
                path = []
                for pth in reversed(self.stack.stack[:-1]):
                    if pth.schema_element_type == "property":
                        path.insert(0, pth.key)
                        candidate = self._get_schema_class_candidate(
                            schema_class,
                            fingerprint,
                            prefix="".join(x.title() for x in path),
                            known_classes=known_classes,
                        )
                        if candidate:
                            return candidate
                for i in range(1, 100):
                    candidate = self._get_schema_class_candidate(
                        schema_class,
                        fingerprint,
                        suffix=f"_{i}",
                        known_classes=known_classes,
                    )
                    if candidate:
                        return candidate

                raise InvalidModelException(
                    f"Too many marshmallow classes with name {schema_class}. Please specify your own class names"
                )

        return schema_class

    def _get_schema_class_candidate(
        self, schema_class, fingerprint, suffix="", prefix="", known_classes=None
    ):
        print(" ... trying", schema_class, prefix, suffix)
        pkg_clz = schema_class.rsplit(".", maxsplit=1)
        if len(pkg_clz) > 1:
            pkg, clz = f"{pkg_clz[0]}.", pkg_clz[1]
        else:
            pkg = ""
            clz = pkg_clz[0]

        candidate = f"{pkg}{prefix}{clz}{suffix}".rsplit(".", maxsplit=1)
        if len(candidate) > 1:
            candidate = candidate[0] + "." + convert_name_to_python_class(candidate[1])
        else:
            candidate = convert_name_to_python_class(candidate[0])
        if candidate not in known_classes:
            return candidate
        if fingerprint == known_classes[candidate]:
            return candidate
        return None

    def _get_class_name(self, package_name: str, class_name: str):
        if "." not in class_name:
            return f"{package_name}.{class_name}"
        if class_name.startswith("."):
            package_path = package_name.split(".")
            while class_name.startswith("."):
                if package_path:
                    package_path = package_path[:-1]
                class_name = class_name[1:]
            if package_path:
                class_name = f"{'.'.join(package_path)}.{class_name}"
        return class_name

    def facet(self, key, definition={}, props_num=None, create=True):
        key = definition.get("key", key)
        field = definition.get("field", "TermFacet")
        if "searchable" in definition:
            for d in self.stack.top.data.properties:
                self.stack.top.data.properties[d]["facets"] = {
                    "searchable": definition["searchable"]
                }
        if not searchable(definition, create):
            return False
        if props_num == 0:
            return False

        return {"path": key, "class": field, "props_num": props_num}


class NestedDataType(ObjectDataType):
    schema_type = "object"
    mapping_type = "nested"
    marshmallow_field = "ma_fields.Nested"
    model_type = "nested"

    def facet(self, key, definition={}, props_num=None, create=True):
        if "searchable" in definition:
            for d in self.stack.top.data.properties:
                self.stack.top.data.properties[d]["facets"] = {
                    "searchable": definition["searchable"]
                }
        if not searchable(definition, create):
            return False
        upper = self.stack[-2]
        if key in definition:
            key = definition["key"]
        elif upper.json_schema_type == "array":
            key = upper.key
        return {"path": key, "class": "NestedLabeledFacet", "props_num": props_num}


class FlattenDataType(DataType):
    schema_type = "object"
    mapping_type = "flatten"
    marshmallow_field = "ma_fields.Raw"
    model_type = "flatten"

    def facet(self, key, definition={}, props_num=None, create=True):
        if not searchable(definition, create):
            return False
        facet_def = {}
        if "field" in definition:
            key = definition.get("key", key)
            facet_def = {"path": key, "class": definition["field"]}
            facet_def["defined_class"] = True
            return facet_def
        else:
            return False


class ArrayDataType(DataType):
    schema_type = "array"
    mapping_type = None
    marshmallow_field = "ma_fields.List"
    model_type = "array"

    class ModelSchema(DataType.ModelSchema):
        items = fields.Nested(
            lambda: model_validator.validator_class("array-items", strict=False)()
        )
        uniqueItems = fields.Boolean(required=False)
        minItems = fields.Integer(required=False)
        maxItems = fields.Integer(required=False)

    def facet(self, key, definition={}, props_num=None, create=True):
        if "searchable" in definition:
            if "properties" in self.stack.top.data["items"]:
                self.stack.top.data["items"]["facets"] = {
                    "searchable": definition["searchable"]
                }
                for d in self.stack.top.data["items"].properties:
                    self.stack.top.data["items"].properties[d]["facets"] = {
                        "searchable": definition["searchable"]
                    }
            else:
                for d in self.stack.top.data["items"]:
                    self.stack.top.data["items"][d]["facets"] = {
                        "searchable": definition["searchable"]
                    }
        if not searchable(definition, create):
            return False
        key = definition.get("key", key)
        if key not in definition and "keyword" in definition:
            key = key + "_keyword"

        field = definition.get("field", "TermsFacet(field = ")
        facet_def = {}
        if "nested" in definition:
            field = "NestedLabeledFacet"
        elif "field" in definition:
            facet_def["defined_class"] = True
        facet_def["path"] = key
        facet_def["class"] = field

        if props_num == 0:
            return False
        if int(props_num or 0) > 1:
            facet_def["props_num"] = props_num

        return facet_def



from oarepo_model_builder.builders.jsonschema import JSONSchemaBuilder
from oarepo_model_builder.builders.mapping import MappingBuilder
from oarepo_model_builder.datatypes import datatypes
from oarepo_model_builder.invenio.invenio_record_schema import (
    InvenioRecordSchemaBuilder,
)
from oarepo_model_builder.property_preprocessors import PropertyPreprocessor, process
from oarepo_model_builder.stack import ModelBuilderStack
from oarepo_model_builder.utils.deepmerge import deepmerge


class DataTypePreprocessor(PropertyPreprocessor):
    TYPE = "datatype"

    @process(
        model_builder=JSONSchemaBuilder,
        path="/properties/**",
        condition=lambda current, stack: stack.top.schema_element_type
        in ("property", "items"),
    )
    def modify_jsonschema(self, data, stack: ModelBuilderStack, **kwargs):
        datatype = datatypes.get_datatype(
            data, stack.top.key, self.schema.current_model, self.schema, stack
        )
        if not datatype:
            return data
        self.merge_with_data(data, datatype.model_schema())
        return datatype.json_schema()

    @process(
        model_builder=MappingBuilder,
        path="/properties/**",
        condition=lambda current, stack: stack.top.schema_element_type
        in ("property", "items"),
    )
    def modify_mapping(self, data, stack: ModelBuilderStack, **kwargs):
        datatype = datatypes.get_datatype(
            data, stack.top.key, self.schema.current_model, self.schema, stack
        )
        if not datatype:
            return data
        self.merge_with_data(data, datatype.model_schema())
        return datatype.mapping()

    @process(
        model_builder=InvenioRecordSchemaBuilder,
        path="/properties/**",
        condition=lambda current, stack: stack.top.schema_element_type
        in ("property", "items"),
    )
    def modify_marshmallow(self, data, stack: ModelBuilderStack, **kwargs):
        datatype = datatypes.get_datatype(
            data, stack.top.key, self.schema.current_model, self.schema, stack
        )
        if not datatype:
            return data
        self.merge_with_data(data, datatype.model_schema())
        ret = datatype.marshmallow()
        imports = datatype.imports()
        ret.setdefault("imports", []).extend(
            [
                {"import": i.import_path, "alias": i.alias}
                if i.alias
                else {"import": i.import_path}
                for i in imports
            ]
        )

    @process(
        model_builder="*",
        path="/properties/**",
        condition=lambda current, stack: stack.top.schema_element_type
        in ("property", "items"),
        priority=-100,  # fallback priority
    )
    def modify_generic(
        self, data, stack: ModelBuilderStack, output_builder_type, **kwargs
    ):
        datatype = datatypes.get_datatype(
            data, stack.top.key, self.schema.current_model, self.schema, stack
        )
        if not datatype:
            return data
        self.merge_with_data(data, datatype.model_schema())
        method = getattr(datatype, "process_" + output_builder_type, None)
        if method and callable(method):
            return method()
        return data

    def merge_with_data(self, data, extra_schema):
        if extra_schema:
            return deepmerge(data, extra_schema)
        return data



from oarepo_model_builder.templates import templates
from oarepo_model_builder.utils.hyphen_munch import HyphenMunch


def test_load_default_template():
    settings = {
        "python": {
            "templates": {},
        }
    }
    assert templates.get_template("record", HyphenMunch(settings))


def test_load_template_in_settings():
    settings = {"python": {"templates": {"blah": __file__}}}
    assert templates.get_template("blah", HyphenMunch(settings))



import marshmallow as ma
import yaml
from marshmallow import fields
from marshmallow.decorators import PRE_LOAD
from marshmallow.exceptions import ValidationError
from marshmallow_oneofschema import OneOfSchema

from .model_validation import PROPERTY_BY_TYPE_PREFIX, model_validator
from .utils import RegexFieldsSchema


class PropertiesSchema(RegexFieldsSchema):
    class Meta:
        regex_fields = [
            {
                "key": "^.*$",
                "field": lambda: fields.Nested(
                    lambda: model_validator.validator_class("object-field")()
                ),
            },
        ]

    @ma.pre_load(pass_many=False)
    def transform_props_before_load(self, data, **kwargs):
        returned_data = {}
        for k, v in data.items():
            if isinstance(v, str):
                # single-line type followed by yaml => read it
                if "{" not in v:
                    v = {"type": v}
                else:
                    datatype, constraints = v.split("{", maxsplit=1)
                    constraints = constraints.replace(":", ": ")
                    v = {
                        "type": datatype,
                        **yaml.safe_load("blah: {" + constraints)["blah"],
                    }
            if k.endswith("{}"):
                k = k[:-2]
                children, top_props = self.elevate_props(v)
                v = {"type": "object", **top_props, "properties": children}
            if k.endswith("{nested}"):
                k = k[:-8]
                children, top_props = self.elevate_props(v)
                v = {"type": "nested", **top_props, "properties": children}
            if k.endswith("[]"):
                k = k[:-2]
                children, top_props = self.elevate_props(v)
                v = {"type": "array", **top_props, "items": children}
            returned_data[k] = v
        return returned_data

    def elevate_props(self, v):
        children = {}
        top_props = {}
        for k, v in v.items():
            if k[0] == "^":
                top_props[k[1:]] = v
            else:
                children[k] = v
        return children, top_props


class PropertySchemas:
    def __init__(self, is_array=False) -> None:
        self.is_array = is_array

    def get(self, property_type):
        return model_validator.validator_class(
            f"{PROPERTY_BY_TYPE_PREFIX}{property_type}"
        )


class ObjectFieldSchema(OneOfSchema):
    type_field_remove = False
    type_schemas = PropertySchemas(is_array=False)

    def get_obj_type(self, obj):
        return obj["type"]

    def load(self, data, *, many=None, partial=None, unknown=None, **kwargs):
        # OneOfSchema does not call pre-load actions, so add it here explicitly
        if self._has_processors(PRE_LOAD):
            try:
                processed_data = self._invoke_load_processors(
                    PRE_LOAD, data, many=many, original_data=data, partial=partial
                )
            except ValidationError as err:
                errors = err.normalized_messages()
                exc = ValidationError(errors, data=data, valid_data={})
                self.handle_error(exc, data, many=many, partial=partial)
                raise exc
        else:
            processed_data = data

        return super().load(
            processed_data, many=many, partial=partial, unknown=unknown, **kwargs
        )

    @ma.pre_load(pass_many=False)
    def set_type(self, data, **kwargs):
        if not isinstance(data, dict):
            raise ValidationError(f'Must be an object, is "{repr(data)[:20]}..."')
        if "type" not in data:
            if "properties" in data:
                data["type"] = "object"
            if "items" in data:
                data["type"] = "array"
        return data


class ArrayItemsSchema(ObjectFieldSchema):
    type_schemas = PropertySchemas(is_array=True)



from pathlib import Path
from typing import Union

import json5

from oarepo_model_builder.builder import ModelBuilder
from oarepo_model_builder.entrypoints import create_builder_from_entrypoints, load_model
from oarepo_model_builder.fs import InMemoryFileSystem
from oarepo_model_builder.profiles import Profile
from oarepo_model_builder.profiles.extend import ExtendProfile
from oarepo_model_builder.schema import ModelSchema
from oarepo_model_builder.utils.deepmerge import deepmerge


class ModelProfile(Profile):
    def build(
        self,
        model: ModelSchema,
        output_directory: Union[str, Path],
        builder: ModelBuilder,
    ):
        # at first handle "extend"
        if "extend" in model.current_model:
            self.handle_extend(model.current_model.extend, model, builder)
        return super().build(model, output_directory, builder)

    def handle_extend(
        self, extended_schema: str, model: ModelSchema, builder: ModelBuilder
    ):
        """
        extending the model means:
            1. extended schema is read as a schema. If there is a json path behind it,
               it is used to select the model. by default /model is taken
            2. The model will be processed through "extend" profile with
               the output file "model.json5" stored in memory:
                  * properties ending with "-class" on model will be turned to array and suffix changed to -base-classes
                  * jsonschema and mapping will not be touched
                  * marshmallow will have all the properties marked as {read: false, write: false} so that they are not generated
            3. the resulting model will be merged as if use: was used
        """

        extended_schema = extended_schema.split("#", maxsplit=1)
        schema_location = extended_schema[0]
        if len(extended_schema) == 1:
            model_field = "model"
        else:
            model_field = extended_schema[1]
            if model_field.startswith("/"):
                model_field = model_field[1:]
            if "/" in model_field:
                raise ValueError("Can not currently process nested models")

        extended_model = load_model(
            schema_location,
            package=None,
            configs=[],
            black=False,
            isort=False,
            sets=[],
            extra_included=model.included_schemas,
        )
        extended_model.model_field = model_field

        fs = InMemoryFileSystem()
        builder = create_builder_from_entrypoints(
            profile="extend",
            conflict_resolver=builder.conflict_resolver,
            overwrite=builder.overwrite,
            filesystem=fs,
        )

        ExtendProfile().build(extended_model, "", builder)

        loaded_model = json5.loads(fs.read("model.json5"))
        deepmerge(model.current_model, loaded_model)



import sys

import pytest
from marshmallow.exceptions import ValidationError

from oarepo_model_builder.validation.model_validation import model_validator


def test_empty_model_validation():
    model_validator.validate({})
    model_validator.validate({"model": {}})
    model_validator.validate({"model": {"type": "object"}})
    with pytest.raises(ValidationError, match="Bad value.*'object' expected"):
        model_validator.validate({"model": {"type": "blah"}})
    model_validator.validate({"model": {"type": "object", "properties": {}}})
    model_validator.validate({"model": {"type": "object", "properties": None}})


def test_unknown_on_top_validation():
    with pytest.raises(ValidationError, match="'unexpected'.*'Unknown field.'"):
        model_validator.validate({"unexpected": True})


def test_settings_on_model():
    model_validator.validate(
        {
            "model": {
                "profile-package": "test",
                "package-path": "test",
                "jsonschemas-package": "test",
                "mapping-file": "test",
                "collection-url": "test",
                "model-name": "test",
            }
        }
    )


def test_inline_props_on_model():
    assert model_validator.validate(
        {
            "model": {
                "properties": {
                    "a": "boolean",
                    "b": "integer{minimum:1}",
                    "c": "float{exclusiveMaximum: 1.0}",
                    "d": "double",
                }
            }
        }
    ) == {
        "version": "1.0.0",
        "model": {
            "type": "object",
            "properties": {
                "a": {"type": "boolean"},
                "b": {"type": "integer", "minimum": 1},
                "c": {"type": "float", "exclusiveMaximum": 1.0},
                "d": {"type": "double"},
            },
        },
    }



import keyword
import re

from oarepo_model_builder.builders import process
from oarepo_model_builder.utils.jinja import package_name
from oarepo_model_builder.utils.python_name import convert_name_to_python

from ..datatypes import datatypes
from ..outputs.json_stack import JSONStack
from ..utils.deepmerge import deepmerge
from .invenio_base import InvenioBaseClassPythonBuilder

OAREPO_FACETS_PROPERTY = "facets"
OAREPO_SORTABLE_PROPERTY = "sortable"


class InvenioRecordSearchOptionsBuilder(InvenioBaseClassPythonBuilder):
    TYPE = "invenio_record_search"
    class_config = "record-search-options-class"
    template = None

    def begin(self, schema, settings):
        super().begin(schema, settings)
        self.template = "record-search-options"
        self.search_options_data = []
        self.sort_options_data = []
        self.search_options_stack = JSONStack()
        self.facets_definition = []
        self.settings = settings
        if "sortable" in schema.schema:
            self.process_top_sortable(schema.schema["sortable"])
        self.facet_switch = self.schema.model.get("searchable", True)
        self.facet_stack = []

    def finish(self, **extra_kwargs):
        super().finish(
            search_options_data=self.search_options_data,
            facets_definition=self.facets_definition,
            sort_definition=self.sort_options_data,
        )
        python_path = self.class_to_path(self.current_model.record_facets_class)
        self.process_template(
            python_path,
            "record-facets",
            current_package_name=package_name(self.current_model.record_facets_class),
            search_options_data=self.search_options_data,
            **extra_kwargs,
        )

    def process_top_sortable(self, data):
        keys = dir(data)
        for k in keys:
            fields = dir(data[k])
            fields_options = {}
            for field in fields:
                fields_options = deepmerge(fields_options, {field: data[k][field]})
            self.sort_options_data.append({k: fields_options})

    @process("**", condition=lambda current, stack: stack.schema_valid)
    def enter_model_element(self):
        schema_element_type = self.stack.top.schema_element_type
        data = self.stack.top.data
        recurse = True

        if recurse:
            try:
                self.definition = data.get(OAREPO_FACETS_PROPERTY, {})
                fd = datatypes.get_datatype(
                    data, data.type, self.current_model, self.schema, self.stack
                )
                ft = False
                if fd.schema_type == "object":
                    properties = data.get("properties", {})
                    ft = fd.facet(
                        key=self.stack.top.key,
                        props_num=self.properties_types(properties),
                        definition=self.definition,
                        create=self.facet_switch,
                    )
                elif fd.schema_type == "array":
                    ft = fd.facet(
                        key=self.stack.top.key,
                        props_num=self.properties_types(data["items"], True),
                        definition=self.definition,
                        create=self.facet_switch,
                    )
                if ft:
                    self.facet_stack.append(ft)
            except:
                pass

            self.build_children()

        if not self.search_options_stack:
            return

        if schema_element_type == "property":
            sort_definition = data.get(OAREPO_SORTABLE_PROPERTY, None)

            if sort_definition != None:
                self.sort_options_data.append(
                    self.process_sort_options(self.stack.path, sort_definition)
                )

        if schema_element_type == "property" and (
            ("type" in data)
            and (
                datatypes.get_datatype(
                    data, data.type, self.current_model, self.schema, self.stack
                ).schema_type
                != "object"
            )
        ):
            d_type = datatypes.get_datatype(
                data, data.type, self.current_model, self.schema, self.stack
            )
            ft = d_type.facet(
                key=self.stack.top.key,
                definition=self.definition,
                create=self.facet_switch,
            )
            if ft and data.type != "array":
                self.facet_stack.append(ft)
            if len(self.facet_stack) > 0:
                facet_def = ""
                facet_name = ""
                facet_path = ""
                nested_count = 0
                for facet in self.facet_stack:
                    facet_name = (
                        facet_name + convert_name_to_python(facet["path"]) + "_"
                    )
                    facet_path = facet_path + facet["path"] + "."
                    if "defined_class" in facet:
                        facet_def = facet_def + facet["class"]
                    elif facet["class"].startswith("Nested"):
                        nested_count += 1
                        facet_def = (
                            facet_def
                            + f'NestedLabeledFacet(path =" {facet_path[:-1]}", nested_facet='
                        )
                    elif "props_num" in facet:
                        pass
                    else:
                        facet_path = (
                            (facet_path[::-1]).replace(
                                "_keyword."[::-1], ".keyword."[::-1], 1
                            )[::-1]
                            if facet_path.endswith("_keyword.")
                            else facet_path
                        )
                        facet_def = facet_def + facet["class"] + f'"{facet_path[:-1]}"'
                        for i in range(0, nested_count):
                            facet_def = facet_def + ")"
                        facet_def = facet_def + ")"
                self.clean_stack()

                facet_name = facet_name[:-1]
                if facet_def:
                    self.search_options_data.append({facet_name: facet_def})
                    search_ops_name = "facets." + facet_name
                    self.facets_definition.append({facet_name: search_ops_name})

    def process_search_options(self, data, field_class):
        text = ""
        for x in data:
            if text == "":
                text = text + x[0] + ' = "' + x[1] + '"'
            else:
                text = text + ", " + x[0] + ' = "' + x[1] + '"'
        return field_class + "(" + text + ")"

    def clean_stack(self):
        self.facet_stack.reverse()
        del_indices = []
        del self.facet_stack[:1]
        for facet in self.facet_stack:
            if "props_num" in facet and facet["props_num"] == 1:
                del_indices.append(self.facet_stack.index(facet))
            elif "props_num" in facet:
                facet["props_num"] = facet["props_num"] - 1
                break
        for i in del_indices[::-1]:
            del self.facet_stack[i]
        self.facet_stack.reverse()

    def properties_types(self, data, array=False):
        count = 0
        ft = False
        if array:
            if "type" in data and data["type"] == "object":
                self.definition["obj"] = True
                data = data["properties"]
            elif "type" in data and data["type"] == "nested":
                self.definition["nested"] = True
            elif "type" in data and data["type"] == "fulltext+keyword":
                self.definition["keyword"] = True
                return 1
            elif "type" in data:
                fd = datatypes.get_datatype(
                    data, data.type, self.current_model, self.schema, self.stack
                )
                ft = fd.facet(key="")
                if ft:
                    return 1
            else:
                return 0
        for d in data:
            if "properties" in data[d]:
                count = count + 1
            elif "type" in data[d]:
                fd = datatypes.get_datatype(
                    data[d], data[d].type, self.current_model, self.schema, self.stack
                )
                ft = fd.facet(key=d)
                if ft:
                    count = count + 1
        return count

    def process_name(self, path, type):
        path_array = (path.split("/"))[2:]
        name = path_array[0]
        if len(path_array) == 1:
            return name
        path_array.pop(0)
        last_path = ""
        for path in path_array:
            if last_path != "properties" and path == "items":
                continue
            last_path = path
            if path == "properties":
                continue
            if type == "name":
                name = name + "_" + path
            elif type == "field":
                name = name + "." + path

        return name

    def process_sort_options(self, path, definition):
        field = self.process_name(path=path, type="field")

        key = definition.get("key", "")
        if key == "":
            key = self.process_name(path, "name")
        order = definition.get("order", "asc")
        if order == "desc":
            field = "-" + field

        return {key: dict(fields=[field])}



import os
import re

from oarepo_model_builder.entrypoints import create_builder_from_entrypoints, load_model
from oarepo_model_builder.fs import InMemoryFileSystem

DUMMY_YAML = "test.yaml"


def test_include_invenio():
    schema = load_model(
        DUMMY_YAML,
        "test",
        model_content={
            "model": {
                "use": "invenio",
                "properties": {
                    "a": "fulltext+keyword",
                    "b": {
                        "type": "keyword",
                        "facets": {"field": 'TermsFacet(field="cosi")'},
                    },
                    "c": "fulltext",
                },
            },
        },
        isort=False,
        black=False,
    )

    filesystem = InMemoryFileSystem()
    builder = create_builder_from_entrypoints(filesystem=filesystem)
    builder.build(schema, "")

    data = builder.filesystem.open(
        os.path.join("test", "services", "records", "facets.py")
    ).read()
    assert re.sub(r"\s", "", data) == re.sub(
        r"\s",
        "",
        """
\"""Facet definitions.\"""

from invenio_records_resources.services.records.facets import TermsFacet
from invenio_search.engine import dsl
from oarepo_runtime.facets.nested_facet import NestedLabeledFacet



a_keyword = TermsFacet(field = "a.keyword")



b = TermsFacet(field="cosi")



_id = TermsFacet(field = "id")



created = TermsFacet(field = "created")



updated = TermsFacet(field = "updated")



_schema = TermsFacet(field = "$schema")
    """,
    )


def test_nested():
    schema = load_model(
        DUMMY_YAML,
        "test",
        model_content={
            "model": {
                "use": "invenio",
                "properties": {
                    "b": {
                        "type": "nested",
                        "properties": {
                            "c": {
                                "type": "keyword",
                            },
                            "d": {"type": "fulltext+keyword"},
                            "f": {
                                "type": "nested",
                                "properties": {"g": {"type": "keyword"}},
                            },
                        },
                    }
                },
            },
        },
        isort=False,
        black=False,
    )

    filesystem = InMemoryFileSystem()
    builder = create_builder_from_entrypoints(filesystem=filesystem)

    builder.build(schema, "")

    data = builder.filesystem.open(
        os.path.join("test", "services", "records", "facets.py")
    ).read()
    assert re.sub(r"\s", "", data) == re.sub(
        r"\s",
        "",
        """
\"""Facet definitions.\"""

from invenio_records_resources.services.records.facets import TermsFacet
from invenio_search.engine import dsl
from oarepo_runtime.facets.nested_facet import NestedLabeledFacet




b_c = NestedLabeledFacet(path =" b", nested_facet = TermsFacet(field = "b.c"))



b_d_keyword = NestedLabeledFacet(path =" b", nested_facet=TermsFacet(field = "b.d.keyword"))




b_f_g = NestedLabeledFacet(path =" b", nested_facet=NestedLabeledFacet(path =" b.f", nested_facet=TermsFacet(field = "b.f.g")))



_id = TermsFacet(field = "id")



created = TermsFacet(field = "created")



updated = TermsFacet(field = "updated")



_schema = TermsFacet(field = "$schema")
    
""",
    )


def test_object():
    schema = load_model(
        DUMMY_YAML,
        "test",
        model_content={
            "model": {
                "use": "invenio",
                "properties": {
                    "b": {
                        "type": "object",
                        "properties": {
                            "c": {
                                "type": "keyword",
                            },
                            "d": {"type": "fulltext+keyword"},
                            "f": {
                                "type": "object",
                                "properties": {"g": {"type": "keyword"}},
                            },
                            "e": "fulltext",
                        },
                    }
                },
            },
        },
        isort=False,
        black=False,
    )

    filesystem = InMemoryFileSystem()
    builder = create_builder_from_entrypoints(filesystem=filesystem)

    builder.build(schema, "")

    data = builder.filesystem.open(
        os.path.join("test", "services", "records", "facets.py")
    ).read()
    print(data)
    assert re.sub(r"\s", "", data) == re.sub(
        r"\s",
        "",
        """
        \"""Facet definitions.\"""

from invenio_records_resources.services.records.facets import TermsFacet
from invenio_search.engine import dsl
from oarepo_runtime.facets.nested_facet import NestedLabeledFacet




b_c = TermsFacet(field = "b.c")



b_d_keyword = TermsFacet(field = "b.d.keyword")



b_f_g = TermsFacet(field = "b.f.g")



_id = TermsFacet(field = "id")



created = TermsFacet(field = "created")



updated = TermsFacet(field = "updated")



_schema = TermsFacet(field = "$schema")
""",
    )


def test_nest_obj():
    schema = load_model(
        DUMMY_YAML,
        "test",
        model_content={
            "model": {
                "use": "invenio",
                "properties": {
                    "b_nes": {
                        "type": "nested",
                        "properties": {
                            "c": {
                                "type": "keyword",
                            },
                            "d": {"type": "fulltext+keyword"},
                            "f": {
                                "type": "object",
                                "properties": {"g": {"type": "keyword"}},
                            },
                        },
                    },
                    "b_obj": {
                        "type": "object",
                        "properties": {
                            "c": {
                                "type": "keyword",
                            },
                            "d": {"type": "fulltext+keyword"},
                            "f": {
                                "type": "nested",
                                "properties": {"g": {"type": "keyword"}},
                            },
                        },
                    },
                },
            },
        },
        isort=False,
        black=False,
    )

    filesystem = InMemoryFileSystem()
    builder = create_builder_from_entrypoints(filesystem=filesystem)

    builder.build(schema, "")

    data = builder.filesystem.open(
        os.path.join("test", "services", "records", "facets.py")
    ).read()
    assert re.sub(r"\s", "", data) == re.sub(
        r"\s",
        "",
        """
        \"""Facet definitions.\"""

from invenio_records_resources.services.records.facets import TermsFacet
from invenio_search.engine import dsl
from oarepo_runtime.facets.nested_facet import NestedLabeledFacet




b_nes_c = NestedLabeledFacet(path ="b_nes", nested_facet=TermsFacet(field = "b_nes.c"))



b_nes_d_keyword = NestedLabeledFacet(path ="b_nes", nested_facet=TermsFacet(field = "b_nes.d.keyword"))



b_nes_f_g = NestedLabeledFacet(path ="b_nes", nested_facet=TermsFacet(field = "b_nes.f.g"))



b_obj_c = TermsFacet(field = "b_obj.c")



b_obj_d_keyword = TermsFacet(field = "b_obj.d.keyword")



b_obj_f_g = NestedLabeledFacet(path ="b_obj.f", nested_facet=TermsFacet(field = "b_obj.f.g"))



_id = TermsFacet(field = "id")



created = TermsFacet(field = "created")



updated = TermsFacet(field = "updated")



_schema = TermsFacet(field = "$schema")
""",
    )


def test_array():
    schema = load_model(
        DUMMY_YAML,
        "test",
        model_content={
            "model": {
                "use": "invenio",
                "properties": {
                    "a[]": "keyword",
                    "b[]": "fulltext",
                    "c[]": "fulltext+keyword",
                },
            },
        },
        isort=False,
        black=False,
    )

    filesystem = InMemoryFileSystem()
    builder = create_builder_from_entrypoints(filesystem=filesystem)

    builder.build(schema, "")

    data = builder.filesystem.open(
        os.path.join("test", "services", "records", "facets.py")
    ).read()
    assert re.sub(r"\s", "", data) == re.sub(
        r"\s",
        "",
        """
        \"""Facet definitions.\"""

from invenio_records_resources.services.records.facets import TermsFacet
from invenio_search.engine import dsl
from oarepo_runtime.facets.nested_facet import NestedLabeledFacet




a = TermsFacet(field = "a")



c_keyword = TermsFacet(field = "c.keyword")


_id = TermsFacet(field = "id")



created = TermsFacet(field = "created")



updated = TermsFacet(field = "updated")



_schema = TermsFacet(field = "$schema")
""",
    )


def test_array_nested():
    schema = load_model(
        DUMMY_YAML,
        "test",
        model_content={
            "model": {
                "use": "invenio",
                "properties": {
                    "obj": {
                        "type": "object",
                        "properties": {
                            "arr": {
                                "type": "array",
                                "items": {
                                    "type": "nested",
                                    "properties": {
                                        "d": {"type": "keyword"},
                                        "e": {
                                            "type": "object",
                                            "properties": {"f": "keyword"},
                                        },
                                    },
                                },
                            }
                        },
                    },
                },
            },
        },
        isort=False,
        black=False,
    )

    filesystem = InMemoryFileSystem()
    builder = create_builder_from_entrypoints(filesystem=filesystem)

    builder.build(schema, "")

    data = builder.filesystem.open(
        os.path.join("test", "services", "records", "facets.py")
    ).read()
    assert re.sub(r"\s", "", data) == re.sub(
        r"\s",
        "",
        """
        \"""Facet definitions.\"""

from invenio_records_resources.services.records.facets import TermsFacet
from invenio_search.engine import dsl
from oarepo_runtime.facets.nested_facet import NestedLabeledFacet


obj_arr_d = NestedLabeledFacet(path ="obj.arr", nested_facet=TermsFacet(field = "obj.arr.d"))

obj_arr_e_f = NestedLabeledFacet(path ="obj.arr", nested_facet=TermsFacet(field = "obj.arr.e.f"))


_id = TermsFacet(field = "id")



created = TermsFacet(field = "created")



updated = TermsFacet(field = "updated")



_schema = TermsFacet(field = "$schema")
""",
    )


def test_top_facets():
    schema = load_model(
        DUMMY_YAML,
        "test",
        model_content={
            "model": {
                "use": "invenio",
                "searchable": False,
                "properties": {
                    "a": {"type": "fulltext+keyword", "facets": {"searchable": True}},
                    "b": {
                        "type": "keyword",
                        "facets": {"field": 'TermsFacet(field="cosi")'},
                    },
                    "c": "keyword",
                    "arr": {
                        "type": "array",
                        "facets": {"searchable": True},
                        "items": {
                            "type": "nested",
                            "properties": {
                                "d": {"type": "keyword"},
                                "e": {"type": "object", "properties": {"f": "keyword"}},
                            },
                        },
                    },
                    "lst2": {
                        "type": "array",
                        "items": {"type": "keyword"},
                        "facets": {"field": 'TermsFacet(field="cosi")'},
                    },
                    "lst[]": "keyword",
                },
            },
        },
        isort=False,
        black=False,
    )

    filesystem = InMemoryFileSystem()
    builder = create_builder_from_entrypoints(filesystem=filesystem)
    builder.build(schema, "")

    data = builder.filesystem.open(
        os.path.join("test", "services", "records", "facets.py")
    ).read()
    print(data)
    assert re.sub(r"\s", "", data) == re.sub(
        r"\s",
        "",
        """
\"""Facet definitions.\"""

from invenio_records_resources.services.records.facets import TermsFacet
from invenio_search.engine import dsl
from oarepo_runtime.facets.nested_facet import NestedLabeledFacet



a_keyword = TermsFacet(field = "a.keyword")



b = TermsFacet(field="cosi")



arr_d = NestedLabeledFacet(path ="arr", nested_facet=TermsFacet(field = "arr.d"))

arr_e_f = NestedLabeledFacet(path ="arr", nested_facet=TermsFacet(field = "arr.e.f"))

lst2 = TermsFacet(field="cosi")

_id = TermsFacet(field = "id")



created = TermsFacet(field = "created")



updated = TermsFacet(field = "updated")



_schema = TermsFacet(field = "$schema")

    """,
    )


def test_searchable_true():
    schema = load_model(
        DUMMY_YAML,
        "test",
        model_content={
            "model": {
                "use": "invenio",
                "properties": {
                    "a": {"type": "fulltext+keyword", "facets": {"searchable": False}},
                    "b": {
                        "type": "keyword",
                        "facets": {"field": 'TermsFacet(field="cosi")'},
                    },
                    "c": "fulltext",
                    "f": {
                        "type": "object",
                        "facets": {"searchable": False},
                        "properties": {"g": {"type": "keyword"}},
                    },
                },
            },
        },
        isort=False,
        black=False,
    )

    filesystem = InMemoryFileSystem()
    builder = create_builder_from_entrypoints(filesystem=filesystem)
    builder.build(schema, "")
    data = builder.filesystem.open(
        os.path.join("test", "services", "records", "facets.py")
    ).read()
    print(data)

    assert re.sub(r"\s", "", data) == re.sub(
        r"\s",
        "",
        """
\"""Facet definitions.\"""

from invenio_records_resources.services.records.facets import TermsFacet
from invenio_search.engine import dsl
from oarepo_runtime.facets.nested_facet import NestedLabeledFacet




b = TermsFacet(field="cosi")


_id = TermsFacet(field = "id")



created = TermsFacet(field = "created")



updated = TermsFacet(field = "updated")



_schema = TermsFacet(field = "$schema")

    """,
    )



from oarepo_model_builder.stack.schema import model_paths


def path_valid(*path):
    ms = model_paths
    for p in path:
        ms = ms.get(p)
    return ms.valid


def test_is_schema_element():
    assert path_valid("type")
    assert path_valid("$id")
    assert path_valid("$schema")
    assert path_valid("properties")
    assert path_valid("properties", "a")
    assert path_valid("properties", "a", "type")
    assert path_valid("properties", "a", "items")
    assert path_valid("properties", "a", "items", "type")
    assert path_valid("properties", "a", "properties")

    assert not path_valid("test")
    assert not path_valid("blah", "properties", "a", "test")



from pathlib import Path

from oarepo_model_builder.fs import InMemoryFileSystem
from oarepo_model_builder.outputs.jsonschema import JSONSchemaOutput

try:
    import json5
except ImportError:
    import json as json5


def test_create_simple_schema():
    class FakeBuilder:
        filesystem = InMemoryFileSystem()

    output = JSONSchemaOutput(FakeBuilder(), Path("blah.json"))
    output.begin()
    output.enter("properties", {})
    output.enter("a", {})
    output.primitive("type", "string")
    output.leave()
    output.leave()
    output.finish()

    data = json5.load(FakeBuilder.filesystem.open("blah.json"))

    assert data == {"properties": {"a": {"type": "string"}}}



import os
from pathlib import Path
from typing import Dict

from oarepo_model_builder.utils.jinja import split_base_name

from ..schema import ModelSchema
from . import ModelPreprocessor


class BaseClassesModelPreprocessor(ModelPreprocessor):
    TYPE = "extend-base-classes"

    def transform(self, schema: ModelSchema, settings: Dict):
        model = schema.model
        marshmallow = model.pop("marshmallow")
        for k in list(model.keys()):
            if k.endswith("-class"):
                prefix = k[:-6]
                model[f"{prefix}-bases"] = [model.pop(k)]
            elif k not in ("type", "properties"):
                model.pop(k)  # pop all other stuff
        model["marshmallow"] = {"base-classes": [marshmallow["schema-class"]]}



from pathlib import Path


class OutputBase:
    output_type = None

    def __init__(self, builder, path: Path):
        self.builder = builder
        self.path: Path = path
        self.executable = False

    def begin(self):
        raise NotImplemented()

    def finish(self):
        raise NotImplemented()

    @property
    def created(self):
        raise NotImplementedError()

    def make_executable(self):
        self.executable = True



def searchable(definition, create):
    if create:
        if definition != {}:
            if "searchable" in definition and not definition["searchable"]:
                return False
            else:
                return True
        else:
            return True
    else:
        if definition != {}:
            if "searchable" in definition and not definition["searchable"]:
                return False
            else:
                return True
        else:
            return False



import os

from oarepo_model_builder.builder import ModelBuilder
from oarepo_model_builder.builders.mapping import MappingBuilder
from oarepo_model_builder.datatypes import datatypes
from oarepo_model_builder.fs import InMemoryFileSystem
from oarepo_model_builder.model_preprocessors.default_values import (
    DefaultValuesModelPreprocessor,
)
from oarepo_model_builder.outputs.mapping import MappingOutput
from oarepo_model_builder.outputs.python import PythonOutput
from oarepo_model_builder.schema import ModelSchema
from oarepo_model_builder.validation.model_validation import model_validator
from tests.multilang import MultilangPreprocessor, MultilingualDataType, UIValidator

try:
    import json5
except ImportError:
    import json as json5


def test_simple_mapping_builder():
    model = {"properties": {"a": {"type": "keyword", "mapping": {"type": "text"}}}}
    data = build_model(model)

    assert data == {"mappings": {"properties": {"a": {"type": "text"}}}}


def test_array_mapping_builder():
    model = {
        "properties": {
            "a": {
                "type": "array",
                "items": {"type": "keyword", "mapping": {"type": "text"}},
            }
        }
    }
    data = build_model(model)

    assert data == {"mappings": {"properties": {"a": {"type": "text"}}}}


def build_model(model):
    builder = ModelBuilder(
        output_builders=[MappingBuilder],
        outputs=[MappingOutput, PythonOutput],
        model_preprocessors=[DefaultValuesModelPreprocessor],
        filesystem=InMemoryFileSystem(),
    )
    builder.build(
        model=ModelSchema(
            "",
            {
                "settings": {
                    "python": {"use-isort": False, "use-black": False},
                    "opensearch": {"version": "os-v2"},
                },
                "model": {"package": "test", **model},
            },
        ),
        output_dir="",
    )
    data = json5.load(
        builder.filesystem.open(
            os.path.join(
                "test", "records", "mappings", "os-v2", "test", "test-1.0.0.json"
            )
        )
    )
    return data


def test_mapping_preprocessor():
    datatypes._prepare_datatypes()
    if UIValidator not in model_validator.validator_map["property"]:
        model_validator.validator_map["property"].append(UIValidator)
    datatypes.datatype_map["multilingual"] = MultilingualDataType

    builder = ModelBuilder(
        output_builders=[MappingBuilder],
        outputs=[MappingOutput, PythonOutput],
        model_preprocessors=[DefaultValuesModelPreprocessor],
        property_preprocessors=[MultilangPreprocessor],
        filesystem=InMemoryFileSystem(),
    )

    builder.build(
        model=ModelSchema(
            "",
            {
                "settings": {
                    "python": {"use-isort": False, "use-black": False},
                    "opensearch": {"version": "os-v2"},
                },
                "model": {
                    "package": "test",
                    "properties": {"a": {"type": "multilingual"}},
                },
            },
        ),
        output_dir="",
    )

    data = json5.load(
        builder.filesystem.open(
            os.path.join(
                "test", "records", "mappings", "os-v2", "test", "test-1.0.0.json"
            )
        )
    )

    assert data == {
        "mappings": {
            "properties": {
                "a": {
                    "type": "object",
                    "properties": {
                        "lang": {"type": "keyword"},
                        "value": {"type": "text"},
                    },
                },
                "a_cs": {"type": "text"},
            }
        }
    }



import pathlib
import re
import typing

import marshmallow as ma
from marshmallow import fields
from marshmallow.error_store import ErrorStore
from marshmallow.exceptions import ValidationError
from marshmallow_oneofschema import OneOfSchema
from marshmallow_union import Union

from .model_validation import model_validator


class PathOrString(fields.Field):
    def _serialize(self, value, attr, obj, **kwargs):
        return value  # keep the pathlib.Path in here

    def _deserialize(self, value, attr, data, **kwargs):
        if not value:
            return value
        if isinstance(value, pathlib.Path):
            return value
        return value


class ModelDefaults(ma.Schema):
    package = fields.String(required=False)
    profile_package = fields.String(data_key="profile-package", required=False)
    package_path = PathOrString(data_key="package-path")
    jsonschemas_package = fields.String(data_key="jsonschemas-package", required=False)
    mapping_file = fields.String(data_key="mapping-file", required=False)
    collection_url = fields.String(data_key="collection-url", required=False)
    model_name = fields.String(data_key="model-name", required=False)
    schema_name = fields.String(data_key="schema-name", required=False)
    index_name = fields.String(data_key="index-name", required=False)
    script_import_sample_data = fields.String(
        data_key="script-import-sample-data", required=False
    )
    kebap_package = fields.String(data_key="kebap-package", required=False)
    schema_file = fields.String(data_key="schema-file", required=False)
    package_base = fields.String(data_key="package-base", required=False)
    mapping_package = fields.String(data_key="mapping-package", required=False)
    schema_server = fields.String(data_key="schema-server", required=False)
    schema_version = fields.String(data_key="schema-version", required=False)
    package_base_upper = fields.String(data_key="package-base-upper", required=False)
    saved_model_file = PathOrString(data_key="saved-model-file", required=False)
    config_dummy_class = fields.String(data_key="config-dummy-class", required=False)
    config_package = fields.String(data_key="config-package", required=False)
    config_resource_class_key = fields.String(
        data_key="config-resource-class-key", required=False
    )
    config_resource_config_key = fields.String(
        data_key="config-resource-config-key", required=False
    )
    config_service_class_key = fields.String(
        data_key="config-service-class-key", required=False
    )
    config_service_config_key = fields.String(
        data_key="config-service-config-key", required=False
    )
    config_resource_register_blueprint_key = fields.String(
        data_key="config-resource-register-blueprint-key", required=False
    )
    create_blueprint_from_app = fields.String(
        data_key="create-blueprint-from-app", required=False
    )
    ext_class = fields.String(data_key="ext-class", required=False)
    flask_extension_name = fields.String(
        data_key="flask-extension-name", required=False
    )
    cli_function = fields.String(data_key="cli-function", required=False)
    proxies_current_resource = fields.String(
        data_key="proxies-current-resource", required=False
    )
    proxies_current_service = fields.String(
        data_key="proxies-current-service", required=False
    )
    record_class = fields.String(data_key="record-class", required=False)
    record_dumper_class = fields.String(data_key="record-dumper-class", required=False)
    record_facets_class = fields.String(data_key="record-facets-class", required=False)
    record_jsonschemas_setup_cfg = fields.String(
        data_key="record-jsonschemas-setup-cfg", required=False
    )
    record_mapping_setup_cfg = fields.String(
        data_key="record-mapping-setup-cfg", required=False
    )
    record_api_blueprints_setup_cfg = fields.String(
        data_key="record-api-blueprints-setup-cfg", required=False
    )
    record_blueprints_setup_cfg = fields.String(
        data_key="record-blueprints-setup-cfg", required=False
    )
    oarepo_models_setup_cfg = fields.String(
        data_key="oarepo-models-setup-cfg", required=False
    )
    flask_commands_setup_cfg = fields.String(
        data_key="flask-commands-setup-cfg", required=False
    )
    record_metadata_class = fields.String(
        data_key="record-metadata-class", required=False
    )
    record_metadata_table_name = fields.String(
        data_key="record-metadata-table-name", required=False
    )
    record_permissions_class = fields.String(
        data_key="record-permissions-class", required=False
    )
    record_prefix_snake = fields.String(data_key="record-prefix-snake", required=False)
    record_resource_blueprint_name = fields.String(
        data_key="record-resource-blueprint-name", required=False
    )
    record_resource_class = fields.String(
        data_key="record-resource-class", required=False
    )
    record_resource_config_class = fields.String(
        data_key="record-resource-config-class", required=False
    )
    record_schema_class = fields.String(data_key="record-schema-class", required=False)
    record_schema_metadata_alembic = fields.String(
        data_key="record-schema-metadata-alembic", required=False
    )
    record_schema_metadata_class = fields.String(
        data_key="record-schema-metadata-class", required=False
    )
    record_schema_metadata_setup_cfg = fields.String(
        data_key="record-schema-metadata-setup-cfg", required=False
    )
    record_search_options_class = fields.String(
        data_key="record-search-options-class", required=False
    )
    record_service_class = fields.String(
        data_key="record-service-class", required=False
    )
    record_service_config_class = fields.String(
        data_key="record-service-config-class", required=False
    )
    record_prefix = fields.String(data_key="record-prefix", required=False)
    record_records_package = fields.String(
        data_key="record-records-package", required=False
    )
    record_services_package = fields.String(
        data_key="record-services-package", required=False
    )
    record_resources_package = fields.String(
        data_key="record-resources-package", required=False
    )
    config_dummy_bases = fields.List(
        fields.String(), data_key="config-dummy-bases", required=False
    )
    ext_bases = fields.List(fields.String(), data_key="ext-bases", required=False)
    record_facets_bases = fields.List(
        fields.String(), data_key="record-facets-bases", required=False
    )
    record_schema_metadata_bases = fields.List(
        fields.String(), data_key="record-schema-metadata-bases", required=False
    )
    record_bases = fields.List(fields.String(), data_key="record-bases", required=False)
    record_dumper_bases = fields.List(
        fields.String(), data_key="record-dumper-bases", required=False
    )
    record_metadata_bases = fields.List(
        fields.String(), data_key="record-metadata-bases", required=False
    )
    record_permissions_bases = fields.List(
        fields.String(), data_key="record-permissions-bases", required=False
    )
    record_resource_bases = fields.List(
        fields.String(), data_key="record-resource-bases", required=False
    )
    record_resource_config_bases = fields.List(
        fields.String(), data_key="record-resource-config-bases", required=False
    )
    record_schema_bases = fields.List(
        fields.String(), data_key="record-schema-bases", required=False
    )
    record_search_options_bases = fields.List(
        fields.String(), data_key="record-search-options-bases", required=False
    )
    record_service_bases = fields.List(
        fields.String(), data_key="record-service-bases", required=False
    )
    record_service_config_bases = fields.List(
        fields.String(), data_key="record-service-config-bases", required=False
    )
    record_service_config_components = fields.List(
        fields.String(), data_key="record-service-config-components", required=False
    )
    record_dumper_extensions = fields.List(
        fields.String(), data_key="record-dumper-extensions", required=False
    )
    generate_record_pid_field = fields.Boolean(
        data_key="generate-record-pid-field", required=False
    )
    record_service_config_generate_links = fields.Boolean(
        data_key="record-service-config-generate-links", required=False
    )
    invenio_config_extra_code = fields.String(
        data_key="invenio-config-extra-code", required=False
    )
    invenio_ext_extra_code = fields.String(
        data_key="invenio-ext-extra-code", required=False
    )
    invenio_proxies_extra_code = fields.String(
        data_key="invenio-proxies-extra-code", required=False
    )
    invenio_record_extra_code = fields.String(
        data_key="invenio-record-extra-code", required=False
    )
    invenio_record_dumper_extra_code = fields.String(
        data_key="invenio-record-dumper-extra-code", required=False
    )
    invenio_record_facets_extra_code = fields.String(
        data_key="invenio-record-facets-extra-code", required=False
    )
    invenio_record_metadata_extra_code = fields.String(
        data_key="invenio-record-metadata-extra-code", required=False
    )
    invenio_record_object_schema_extra_code = fields.String(
        data_key="invenio-record-object-schema-extra-code", required=False
    )
    invenio_record_permissions_extra_code = fields.String(
        data_key="invenio-record-permissions-extra-code", required=False
    )
    invenio_record_resource_extra_code = fields.String(
        data_key="invenio-record-resource-extra-code", required=False
    )
    invenio_record_resource_config_extra_code = fields.String(
        data_key="invenio-record-resource-config-extra-code", required=False
    )
    invenio_record_schema_extra_code = fields.String(
        data_key="invenio-record-schema-extra-code", required=False
    )
    invenio_record_search_options_extra_code = fields.String(
        data_key="invenio-record-search-options-extra-code", required=False
    )
    invenio_record_service_extra_code = fields.String(
        data_key="invenio-record-service-extra-code", required=False
    )
    invenio_record_service_config_extra_code = fields.String(
        data_key="invenio-record-service-config-extra-code", required=False
    )
    invenio_version_extra_code = fields.String(
        data_key="invenio-version-extra-code", required=False
    )
    invenio_views_extra_code = fields.String(
        data_key="invenio-views-extra-code", required=False
    )
    service_id = fields.String(data_key="service-id", required=False)
    extension_suffix = fields.String(data_key="extension-suffix", required=False)

    mapping = fields.Nested(ma.Schema())
